/* PB0 - Vantec strobe 1
 * PB1 - Vantec strobe 2
 * PB2 - Vantec strobe 3
 * PB3 - Data LED
 * PB4 - Vantec data 0
 * PB5 - Vantec data 1
 *
 * Port C - Analog inputs
 *
 * PD0 - RX
 * PD1 - TX
 * PD2-7 - Vantec data 2-7
 */

#include <string.h>
#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "uart.h"

#define LED_DATA		(1 << 3)

#define DDR_STROBE		DDRB
#define PORT_STROBE		PORTB

#define STROBE_1		(1 << 0)
#define STROBE_2		(1 << 1)
#define STROBE_3		(1 << 2)

/* How many main loop iterations before computer speeds time out */
#define COMP_AGE_LIMIT	50

/* Speeds received from the computer */
uint8_t comp_speed[6];

/* Main loop iterations since the most recent data from the computer */
uint8_t comp_age;

/* Received data */
uint8_t rx_buf[8];

/* The position in rx_buf of the next character to be received */
uint8_t rx_pos;

ISR(UART_RX_vect)
{
	uint8_t ch;
	
	ch = UDR;
	
	if (UCSRA & ((1 << FE) | (1 << DOR)))
	{
		/* Bad or missed data - abort this packet */
		rx_pos = 0;
		return;
	}
	
	if (rx_pos >= sizeof(rx_buf))
	{
		/* Should never happen */
		return;
	}
	
	/* If this is a signature byte, check it */
	if (rx_pos == 0 && ch != 0xc9)
	{
		return;
	}
	
	if (rx_pos == 1 && ch != 0xaa)
	{
		rx_pos = 0;
		return;
	}
	
	/* Store the incoming byte */
	rx_buf[rx_pos++] = ch;
	
	/* Check for finished packet */
	if (rx_pos == 8)
	{
		uint8_t i;
		
		for (i = 0; i < 6; i++)
		{
			comp_speed[i] = rx_buf[i + 2] & 0x3f;
		}
		
		rx_pos = 0;
		comp_age = 0;
	}
}

/* Sets the Vantec data lines */
void set_data(uint8_t data)
{
	/* Bits 0-1 */
	PORTB = (PORTB & 0x0f) | ((data & 0x03) << 4);
	
	/* Bits 2-7 */
	PORTD = data & 0xfc;
}

int main()
{
	uint8_t i;
	
	/* Ports */
	PORTB = 0x20;
	DDRB = 0x3f;
	
	DDRD = 0xfe;
	
	/* UART */
	uart_init();
	rx_pos = 0;
	
	/* ADC */
	ADCSR = (1 << ADEN) | 6;		/* CK/64 */
	
	/* Timer 1 (main loop frequency) */
	TCCR1B = 0x01;					/* CK/1, CTC */
	OCR1H = 0x90;					/* 0x9000 -> 100Hz at 3.6864MHz */
	OCR1L = 0x00;
	
	sei();
	
	/* Initially, no data from the computer */
	comp_age = 0xff;
	
	while (1)
	{
		/* Get most recent computer speeds */
		cli();
		memcpy(comp_speed, comp_speed, 6);
		sei();
		
		/* Time out on computer speeds */
		if (comp_age < COMP_AGE_LIMIT)
		{
			PORTB &= ~LED_DATA;
			comp_age++;
		} else {
			PORTB |= LED_DATA;
			memset(comp_speed, 0, 6);
		}
		
		/* Send speeds */
		set_data(comp_speed[0] | 0x40);		/* Vantec 1 side 1 */
		PORT_STROBE |= STROBE_1;
		PORT_STROBE &= ~STROBE_1;
		
		set_data(comp_speed[1]);			/* Vantec 1 side 2 */
		PORT_STROBE |= STROBE_1;
		PORT_STROBE &= ~STROBE_1;
		
		set_data(comp_speed[2] | 0x40);		/* Vantec 2 side 1 */
		PORT_STROBE |= STROBE_2;
		PORT_STROBE &= ~STROBE_2;
		
		set_data(comp_speed[3]);			/* Vantec 2 side 2 */
		PORT_STROBE |= STROBE_2;
		PORT_STROBE &= ~STROBE_2;
		
		set_data(comp_speed[4] | 0x40);		/* Vantec 3 side 1 */
		PORT_STROBE |= STROBE_3;
		PORT_STROBE &= ~STROBE_3;
		
		set_data(comp_speed[5]);			/* Vantec 3 side 2 */
		PORT_STROBE |= STROBE_3;
		PORT_STROBE &= ~STROBE_3;
		
		/* Sample ADC channels and send status packet */
		uart_send(0x9c);
		uart_send(0x55);
		for (i = 0; i < 6; i++)
		{
			/* Sample this channel */
			ADMUX = i;						/* Select channel */
			ADCSR |= 1 << ADIF;				/* Clear conversion completed flag */
			ADCSR |= 1 << ADSC;				/* Start conversion */
			
			/* Wait for the conversion to complete */
			loop_until_bit_is_set(ADCSR, ADIF);
			
			/* Read the result (order is critical) */
			uart_send(ADCL);
			uart_send(ADCH);
		}
		
		/* Wait until it's time to start the next iteration */
		loop_until_bit_is_set(TIFR, TOV1);
		TIFR = 1 << TOV1;
	}

	return 0;
}
